/* cfuncs.h      complex number functions */

#define fver "v0.02f"
#define fdat "09/10/2008"

#define mver "v0.02f"
#define mdat "07/09/2022"

/* to compensate for floating point precision problems in int routine */
#define INT_COMP 1.0e-8

struct complex
{  double a;
   double b;
};

void ab2rt();
struct complex cx();
int str2cx();
double atofM();

int c_add();
int c_c2f();
int c_c2z();
int c_chs();
int c_conj();
int c_cos();
int c_d2w();
int c_d2v();
int c_db();
int c_div();
int c_exp();
int c_f2c();
int c_f2w();
int c_frac();
int c_imag();
int c_int();
int c_inv();
int c_l2z();
int c_ln(); int c_log(); int c_log10();
int c_mag();
int c_mult();
int c_p2r();
int c_r2p();
int c_real();
int c_res();
int c_s2z();
int c_sin();
int c_sqrt();
int c_sub();
int c_tan();
int c_tenx();
int c_v2d();
int c_v2w();
int c_w2d();
int c_w2f();
int c_w2v();
int c_x2();
int c_YpX();
int c_z2s();
int c_z2c();
int c_z2l();
int c_zs2p();
int c_zpar();

/* convert a+jb to r<t */
void ab2rt(x, r, t)
struct complex x;
double *r, *t;
{  if (x.a==0.0)
   {  *r=x.b; *t=0.0;
      if(x.b<0.0)
      {  *r=-*r; *t=-acos(-1.0)/2.0;
      }
      else
      if (x.b>0) *t=acos(-1.0)/2.0;
   }
   else
   {  *r=sqrt(x.a*x.a+x.b*x.b);
      *t=atan(x.b/x.a); if (x.a<0.0) *t=*t+acos(-1.0);
   }
}

struct complex cx(a,b)
double a,b;
{  struct complex x;

   x.a=a; x.b=b;
   return x;
}

/* a string of the form a+jb or a<b is returned as a complex structure */
/* added hex input 2003/04/30 */
int str2cx(in,z,rd)
char *in;
struct complex *z;
double rd;
{  int x, y, pa, pj, err;
   double r,t,s;
   char in2[1024], *h;

   if(rd<1.0) return 1;
   strcpy(in2, in); /* save */
   y=0; /* to mark location of numeric entry */

   /* look for complex delimiter, J or < */
   pa=-1; pj=-1; err=1;
   for(x=0; in[x]!=0 && (pa+pj)==-2; x++)
   {  if(in[x]=='<') pa=x;
      if(in[x]=='J' || in[x]=='j') pj=x;
   }
   /* Enter complex number as r<theta */
   if (pa!=-1)
   {  if(pa==0) return 1;      /* don't allow < without an r */
      in[pa]=0; r=atofM(in);
      if(in[0]!=0) return 1;   /* non numeric after r entry */
      strcpy(in, in+pa+1);     /* allign to theta entry */
      t=atofM(in)/rd;
      *z=cx(r*cos(t),r*sin(t));
      err=0;
   }
   else
   /* Enter complex number as a+jb */
   {  if (pj!=-1)
      {  if(pj==0) { z->a=0.0; s=1.0; }                           /* no a */
         else
         {  if(in[pj-1]=='-') s=-1.0; else s=1.0;         /* complex sign */
            in[pj-1]=0;
            z->a=atofM(in);
            if(in[0]!=0) return 1;         /* non numeric after "a" entry */
         }
         strcpy(in, in+pj+1);                         /* align to b entry */
         if(in[0]!='0' && in[0]!='.') x=1; else x=0;    /* lone zero flag */
         z->b=atofM(in)*s;
         if(x==1 && z->b==0.0) z->b=1.0*s;               /* allows lone j */
         err=0;
      }
      else /* no J, real only */
      {  /* check for hex, oct, bin in the form 0xN, 0oN, 0bN */
         upit(in2);
         x=16; h=strstr(in2, "0X");
         if(h==NULL) {x=8; h=strstr(in2, "0O");}
         if(h==NULL) {x=2; h=strstr(in2, "0B");}
         if(h!=NULL)
         {  /* convert to decimal */
            for(r=0.0, h=h+2; (*h>47 && *h<58) || (*h>64 && *h<71); h++)
            {  if(*h>64) t=(double)*h-55.0; else t=(double)*h-48.0;
               r=r*(double)x+t;
            }
            *z=cx(r,0.0);
            strcpy(in, h);
            return 0;
         }
         else
         {  *z=cx(atofM(in),0.0); err=0;
             if (z->a==0.0)
               for(x=0; in[x]!=0; x++)
                  if (in[x]!='0' && in[x]!='.' && in[x]!='+' && in[x]!='-' && in[x]!='E')
                     err=1;
         }
      }
   }
   return err;
}

/* atof which allows metric suffixes T, G, M, k, h, c, m, u, p, f, a */
/* deka and deci are not supported because of ambiquity (d?, da?) */
/* anything to the right of the number is returned in str */
double atofM(str)
char *str;
{  int x, n;
   char c;
   double f;

   /* look for 1st part of string that is non numeric */
   for(x=0, n=1, f=1.0; str[x]!=0 && n==1; x++)
   {  c=str[x]; n=0;
      if((c>47 && c<58) || c=='+' || c=='-' || c=='.' || c=='e' || c=='E') n=1;
   }
   if(n==0) /* non numeric found */
   {  switch(c)
      {  case 'T': f=1.0e12;  break;
         case 'G': f=1.0e9;   break;
         case 'M': f=1.0e6;   break; 
         case 'k': f=1.0e3;   break;
         case 'h': f=1.0e2;   break;
         case 'c': f=1.0e-2;  break;
         case 'm': f=1.0e-3;  break;
         case 'u': f=1.0e-6;  break;
         case 'n': f=1.0e-9;  break;
         case 'p': f=1.0e-12; break;
         case 'f': f=1.0e-15; break;
         case 'a': f=1.0e-18; break;
      }
      if(f!=1.0) {str[x-1]=32; x++;}
      f=f*atof(str);
      strcpy(str, str+x-1);
   } 
   else
   {  f=f*atof(str);
      /* allow + or - or e (E) on end of string to remain */
      if(str[x-1]=='+' || str[x-1]=='-' || str[x-1]=='e' || str[x-1]=='E') x--;
      strcpy(str, str+x);
   }
   return f;
}

/* complex add x=x-y */
int c_add(x,y)
struct complex *x,y;
{  x->a = x->a + y.a; x->b = x->b + y.b; return 0;
}

/*? Celcius to Farenheit, x=x*9/5+32*/
/*| c2f|err=c_c2f(xp);|*/
int c_c2f(x)
struct complex *x;
{  double r, t;

   ab2rt(*x, &r, &t);
   r=r*9.0/5.0+32.0;
   *x=cx(r*cos(t),r*sin(t));
   return 0;
}

/* capacitance to impedance z=1/(jwc)*/
int c_c2z(cap, w, er)
struct complex *cap, w;
char *er;
{  double r,t,r1,t1;

   ab2rt(*cap, &r,&t); ab2rt(w,&r1,&t1);
   if(r==0.0 || r1==0.0) { strcpy(er, "Divide by zero"); return 1; }
   r=1.0/(r*r1); t=-(t+t1);
   *cap=cx(r*sin(t),-r*cos(t)); /* this multiplies by -j */
   return 0;
}

/*? Change sign, x=-x*/
/*| chs|err=c_chs(xp);|*/
int c_chs(x)
struct complex *x;
{  x->a=-x->a; x->b=-x->b; return 0;
}

/*? Complex conjucate, x=conj(x)*/
/*| conj|err=c_conj(xp);|*/
int c_conj(x)
struct complex *x;
{  x->b = -x->b; return 0;
}

/*? Cosine, x=cos(x)*/
/*| cos|err=c_cos(xp, rd);|*/
int c_cos(x, rd)
struct complex *x;
double rd;
{  double sa, ca, sb, cb;

   sa=sin(x->a/rd); ca=cos(x->a/rd);
   sb=-(exp(-x->b/rd)-exp(x->b/rd))/2.0;
   cb=(exp(-x->b/rd)+exp(x->b/rd))/2.0;

   *x=cx(ca*cb,-sa*sb); return 0;
}

/*? dBm to watts, x=10^(x/10)*0.001*/
/*| d2w|err=c_d2w(xp);|*/
int c_d2w(x)
struct complex *x;
{  /* double r, t; */

/* This doesn't work with negative numbers,
   I guess I'll do only real
   ab2rt(*x, &r, &t);
   r=pow(10.0,r/10.0)*0.001;
   *x=cx(r*cos(t),r*sin(t)); return 0;
*/
   x->a=pow(10.0,x->a/10.0)*0.001; x->b=0.0;
   return 0;
}

/* complex dBm to volts, p=sqrt(10^(p/10)*0.001*ref) */
int c_d2v(p, ref, er)
struct complex *p, ref;
char *er;
{  double r, t, rr, tt;

   if(ref.a==0.0 && ref.b==0.0)
   { strcpy(er, "Ref=0"); return 1; }
   p->a=p->a/10.0; p->b=p->b/10.0;
   c_tenx(p);
   ab2rt(*p, &r,&t);
   ab2rt(ref, &rr, &tt);
   r=sqrt(r*0.001*rr); t=(t+tt)/2.0;
   *p=cx(r*cos(t),r*sin(t));
   return 0;
}

/* dB [p=10*log10(p/ref)] */
/*? dB, y=10*log10(y/x), pop stack*/
/*| db|err=c_db(yp,*xp,erm);s_pop(sp);|*/
int c_db(p, ref, er)
struct complex *p, ref;
char *er;
{  double r,t, rr,tt;

   ab2rt(*p,&r,&t); ab2rt(ref,&rr,&tt);
   if(r==0.0 && rr==0.0) return 0;
   if(r<=0.0 || rr<=0.0) { strcpy(er, "Log"); return 1; }

   r=10.0*log10(r/rr); t=t-tt;
   *p=cx(r*cos(t),r*sin(t));
   return 0;
}

/* complex divide, y=y/x */
/*? y=x/y, pop stack*/
/*| /|err=c_div(yp,*xp);s_pop(sp);|*/
int c_div(x, y, er)
struct complex *x, y;
char *er;
{  double r,t,r1,t1;

   if (y.a==0.0 && y.b==0.0)
   {  strcpy(er, "Divide by zero"); return 1;
   }
   ab2rt(*x,&r,&t); ab2rt(y,&r1,&t1);
   r=r/r1; t=t-t1;
   *x=cx(r*cos(t),r*sin(t)); return 0;
}

/* complex exp() */
/*? x=exp(x)*/
/*| exp|err=c_exp(xp);|*/
int c_exp(x)
struct complex *x;
{  double r;

   r=exp(x->a);
   x->a=r*cos(x->b); x->b=r*sin(x->b); return 0;
}
 
/*? Farenheit to Celcius, x=(x-32)*5/9*/
/*| f2c|err=c_f2c(xp);|*/
int c_f2c(x)
struct complex *x;
{  double r, t;

   ab2rt(*x, &r, &t);
   r=(r-32.0)*5.0/9.0;
   *x=cx(r*cos(t),r*sin(t));
   return 0;
}

/*? Hertz to radians, x=x*2pi*/
/*| f2w|err=c_f2w(xp);|*/
int c_f2w(x)
struct complex *x;
{  double s;

   s=2.0*acos(-1.0);
   x->a=x->a*s; x->b=x->b*s; return 0;
}

/*? Fractional part of complex number*/
/*| frac|err=c_frac(xp);|*/
int c_frac(x)
struct complex *x;
{  double r, t;

   ab2rt(*x, &r, &t);
   r=r-floor(r+INT_COMP);
   *x=cx(r*cos(t),r*sin(t));
   return 0;
}

/*? Imaginary part of complex number*/
/*| imag|err=c_imag(xp);|*/
int c_imag(x)
struct complex *x;
{  x->a=0.0; return 0;
}

/*? Integer part of complex number*/
/*| int|err=c_int(xp);|*/
int c_int(x)
struct complex *x;
{  double r, t;

   ab2rt(*x, &r, &t);
   r=floor(r+INT_COMP);
   *x=cx(r*cos(t),r*sin(t));
   return 0;
}

/*? Inverse, x=1/x*/
/*| inv|err=c_inv(xp,erm);|*/
int c_inv(x, er)
struct complex *x;
char *er;
{  double r, t;

   if(x->a==0.0 && x->b==0.0)
   {  strcpy(er, "Divide by zero"); return 1;
   }
   ab2rt(*x, &r, &t);
   r=1.0/r; t=-t;
   *x=cx(r*cos(t),r*sin(t));
   return 0;
}

/* inductance to impedance, z=jwL */
int c_l2z(L, w)
struct complex *L, w;
{  double r,t,r1,t1;

   ab2rt(*L, &r,&t); ab2rt(w,&r1,&t1);
   r=r*r1; t=t+t1;    
   *L=cx(-r*sin(t),r*cos(t)); /* this does * +j */
   return 0;
}

/*? Natural logrithm, x=ln(x)*/
/*| ln|err=c_ln(xp,erm);|*/
int c_ln(x, er)
struct complex *x;
char *er;
{  double r, t;

   ab2rt(*x, &r, &t);
   if(r<=0.0) { strcpy(er, "Log"); return 1; }
   *x=cx(log(r),t);
   return 0;
}

/* complex logrithm, if base=0, then e is assumed (ln) */
/*? Logrithm, y=log(y) base x, pop stack*/
/*| log|err=c_log(yp,*xp,erm);s_pop(sp);|*/
int c_log(x, base, er)
struct complex *x, base;
char *er;
{  double r, t, rb, tb;

   ab2rt(*x, &r, &t); ab2rt(base, &rb, &tb);
   if(r<=0.0 || rb<0.0)
   {  strcpy(er, "Log"); return 1;
   }
   if(rb>0.0) rb=log(rb); else rb=1.0;
   *x=cx(log(r)/rb,t/rb);
   return 0;
}

/*? Log base 10, x=log10(x)*/
/*| log10|err=c_log10(xp,erm);|*/
int c_log10(x, er)
struct complex *x;
char *er;
{  double r, t;

   ab2rt(*x, &r, &t);
   if(r<=0.0) { strcpy(er, "Log"); return 1; }
   *x=cx(log(r)/log(10.0), t/log(10.0));
   return 0;
}

/*? Magnitude, x=mag(x)*/
/*| mag|err=c_mag(xp);|*/
int c_mag(x)
struct complex *x;
{  x->a = sqrt(x->a * x->a + x->b * x->b); x->b=0.0; return 0;
}

/* complex multiply, y=x*y */
/*? y=x*y, pop stack*/
/*| *|err=c_mult(yp,*xp);s_pop(sp);|*/
int c_mult(x, y)
struct complex *x, y;
{  double r,t,r1,t1;

   ab2rt(*x,&r,&t); ab2rt(y,&r1,&t1);
   r=r*r1; t=t+t1;
   *x=cx(r*cos(t),r*sin(t)); return 0;
}

/*? Peak to rms, x=x/sqrt(2)*/
/*| p2r|err=c_p2r(xp);|*/
int c_p2r(x)
struct complex *x;
{  *x=cx(x->a/sqrt(2.0), x->b/sqrt(2.0)); return 0;
}

/*? Rms to peak, x=x*sqrt(2)*/
/*| r2p|err=c_r2p(xp);|*/
int c_r2p(x)
struct complex *x;
{  *x=cx(x->a*sqrt(2.0), x->b*sqrt(2.0)); return 0;
}

/*? Real part of complex number*/
/*| real|err=c_real(xp);|*/
int c_real(x)
struct complex *x;
{  x->b=0.0; return 0;
}

/* resonant component */
int c_res(z1, w, er)
struct complex *z1, w;
char *er;
{  struct complex a;
   int err;

   a=w; c_x2(&a); c_mult(&a, *z1, er);
   if((err=c_inv(&a,er))) return err;
   *z1=a; return 0;
}

/* s11 to impedance z=(1+s)/(1-s)*ref */
int c_s2z(s, ref, er)
struct complex *s, ref;
char *er;
{  double r0,t0, r, t, rr, tt;

   ab2rt(ref, &rr, &tt);
   if(rr==0.0) { strcpy(er, "Ref=0"); return 1; }
   ab2rt(cx(1.0+s->a,s->b), &r0, &t0);
   ab2rt(cx(1.0-s->a,-s->b), &r, &t);
   r=r0/r*rr; t=t0-t+tt;
   *s=cx(r*cos(t),r*sin(t));
   return 0;
}

/*? Sine, x=sin(x)*/
/*| sin|err=c_sin(xp, rd);|*/
int c_sin(x, rd)
struct complex *x;
double rd;
{  double sa, ca, sb, cb;

   sa=sin(x->a/rd); ca=cos(x->a/rd);
   sb=-(exp(-x->b/rd)-exp(x->b/rd))/2.0;
   cb=(exp(-x->b/rd)+exp(x->b/rd))/2.0;

   *x=cx(sa*cb,ca*sb); return 0;
}

/*? Square root, x=sqrt(x)*/
/*| sqrt|err=c_sqrt(xp);|*/
int c_sqrt(x)
struct complex *x;
{  double r, t;

   ab2rt(*x, &r, &t);
   r=sqrt(r); t=t/2.0;
   *x=cx(r*cos(t),r*sin(t)); return 0;
}

/* complex subtract x=x-y */
int c_sub(x,y)
struct complex *x,y;
{  x->a = x->a - y.a; x->b = x->b - y.b; return 0;
}

/*? Tangent, x=tan(x)*/
/*| tan|err=c_tan(xp,rd,erm);|*/
int c_tan(x, rd, er)
struct complex *x;
double rd;
char *er;
{  double sa, ca, sb, cb, an, bn, ad, bd, r, t;

   sa=sin(x->a/rd); ca=cos(x->a/rd);
   sb=-(exp(-x->b/rd)-exp(x->b/rd))/2.0;
   cb=(exp(-x->b/rd)+exp(x->b/rd))/2.0;

   an=sa*cb; bn=ca*sb; ad=ca*cb; bd=-sa*sb;
   if(ad==0.0 && bd==0.0)
   {  strcpy(er, "Tangent"); return 1; }
   else
   {  r=sqrt(an*an+bn*bn)/sqrt(ad*ad+bd*bd);
      t=atan(bn/an)-atan(bd/ad);
      *x=cx(r*cos(t),r*sin(t));
   }
   return 0;
}

/* complex 10^x */
int c_tenx(x)
struct complex *x;
{  double r,t;

   r=pow(10.0, x->a); t=x->b*log(10.0);
   *x=cx(r*cos(t),r*sin(t)); return 0;
}

/* complex volts to dBm x=10*log10(x^2/ref/0.001) */
int c_v2d(x, ref, er)
struct complex *x, ref;
char *er;
{  double r, t, rr, tt;

   if(ref.a==0.0 && ref.b==0.0) { strcpy(er, "Ref=0"); return 1; }
   ab2rt(*x, &r, &t); ab2rt(ref, &rr, &tt);
   r=r*r/rr/0.001; t=t+t-tt;
   *x=cx(log(r)/log(10.0)*10.0, t/log(10.0)*10.0); /*10log10(r<t)*/
   return 0;
}

/* volts to watts */
int c_v2w(x, ref, er)
struct complex *x, ref;
char *er;
{  double r, t, rr, tt;

   if(ref.a==0.0 && ref.b==0.0) { strcpy(er, "Ref=0"); return 1; }
   ab2rt(*x, &r, &t); ab2rt(ref, &rr, &tt);
   r=r*r/rr; t=t+t-tt;
   *x=cx(r*cos(t),r*sin(t));
   return 0;
}

/*? Watts to dBm, x=10log10(x/0.001)*/
/*| w2d|err=c_w2d(xp,erm);|*/
int c_w2d(x,er)
struct complex *x;
char *er;
{  double r,t;

   ab2rt(*x, &r, &t);
   if(r<=0.0)
   {  strcpy(er,"Invalid conversion"); return 1;
   }
   *x=cx(10.0*log10(r/0.001), t*10.0); return 0;
}

/*? Radians to Hertz, x = x/(2pi)*/
/*| w2f|err=c_w2f(xp);|*/
int c_w2f(x)
struct complex *x;
{  double s;

   s=2.0*acos(-1);
   *x=cx(x->a/s, x->b/s); return 0;
}

/* complex watts to volts */
int c_w2v(x, ref, er)
struct complex *x, ref;
char *er;
{  double r, t, rr, tt;

   if(ref.a==0.0 && ref.b==0.0) { strcpy(er, "Ref=0"); return 1; }
   ab2rt(*x, &r, &t); ab2rt(ref, &rr, &tt);
   r=sqrt(r*rr); t=(t+tt)/2;
   *x=cx(r*cos(t),r*sin(t));
   return 0;
}

/* complex x squared */
/*? x = x^2*/
/*| x^2|err=c_x2(xp);|*/
int c_x2(x)
struct complex *x;
{  double r,t;

   ab2rt(*x, &r, &t);
   r=r*r; t=t+t;
   *x=cx(r*cos(t),r*sin(t)); return 0;
}

/* Y^X, y=y^x */
/*? y=y^x, pop stack*/
/*| y^x|err=c_YpX(yp,*xp);s_pop(sp);|*/
int c_YpX(y, x)
struct complex *y, x;
{  double r, t, r1, t1;

   if(y->a==0.0 && y->b==0.0) { /* do nothing */ }
   else
   {  ab2rt(*y, &r, &t);
      y->a=log(r); y->b=t;
      ab2rt(*y, &r, &t);
      ab2rt(x, &r1, &t1);
      r=r*r1, t=t+t1;
      y->a=r*cos(t); y->b=r*sin(t);
      r=exp(y->a); t=y->b;
      *y=cx(r*cos(t),r*sin(t));
   }
   return 0;
}

/* impedance to capacitance c=1/(jwz)*/
int c_z2c(z, w, er)
struct complex *z, w;
char *er;
{  double r,t,r1,t1;

   ab2rt(*z, &r,&t); ab2rt(w,&r1,&t1);
   if(r==0.0 || r1==0.0) { strcpy(er, "Divide by zero"); return 1; }
   r=1.0/(r*r1); t=-(t+t1);
   *z=cx(r*sin(t),-r*cos(t)); /* this multiplies by -j */
   return 0;
}

/* impedance to inductance L=z/(jw) */
int c_z2l(z, w, er)
struct complex *z, w;
char *er;
{  double r,t,r1,t1;

   ab2rt(*z, &r,&t); ab2rt(w,&r1,&t1);
   if(r1==0.0) { strcpy(er, "Divide by zero"); return 1; }
   r=r/r1; t=t-t1;
   *z=cx(r*sin(t),-r*cos(t)); /* this multiplies by -j */

   return 0;
}

/* impedance to s11 , s=(z/ref-1)/(z/ref+1) */
int c_z2s(z, ref, er)
struct complex *z, ref;
char *er;
{  double r0, t0, r, t;

   ab2rt(ref, &r0, &t0);
   if(r0==0.0) { strcpy(er, "Ref=0"); return 1; }
   ab2rt(*z, &r, &t);
   r=r/r0; t=t-t0;
   z->a=r*cos(t); z->b=r*sin(t); /* this is z/ref */
   ab2rt(cx(z->a-1.0,z->b), &r0, &t0);
   ab2rt(cx(z->a+1.0,z->b), &r, &t);
   r=r0/r; t=t0-t;
   *z=cx(r*cos(t),r*sin(t));
   return 0;
}

/*? Series impedance to parallel, push, y=Rp, x=jXp*/
/*| s2p|s_push(sp);err=c_zs2p(yp,xp);|*/
int c_zs2p(y,x)
struct complex *y, *x;
{  double r, t;

   if(x->a==0.0 && x->b==0.0)
   {  *y=cx(0.0,0.0); 
   }
   else
   {  ab2rt(*x, &r, &t); r=1.0/r; t=-t;
      *x=cx(0.0           ,-1.0/(r*sin(t)));
      *y=cx(1.0/(r*cos(t)),0.0            );
   }
   return 0;
}

/*? Parallel impedances, y=1/(1/x+1/y), pop*/
/*| par|err=c_zpar(yp,*xp);s_pop(sp);|*/
int c_zpar(x,y)
struct complex *x, y;
{  double r, t, r1, t1;

   if((x->a==0.0 && x->b==0.0) || (y.a==0.0 && y.b==0.0) )
   {  *x=cx(0.0,0.0);
   }
   else
   {  ab2rt(*x, &r, &t); r=1.0/r; t=-t;
      ab2rt(y, &r1, &t1); r1=1.0/r1; t1=-t1;
      x->a=r*cos(t)+r1*cos(t1); x->b=r*sin(t)+r1*sin(t1);
      ab2rt(*x, &r, &t); r=1.0/r; t=-t;
      *x=cx(r*cos(t),r*sin(t));
   }
   return 0;
}
