/* cstack.h          complex stack and memory functions */



#define sver "v0.02a"

#define sdat "07/21/2003"



struct stack

{  int len;

   struct complex d[1];

};



struct mem_element

{  char name[20];

   struct complex d;

};



struct memory

{  int len;

   int max;

   struct mem_element e[1];

};



int m_clear();

struct memory *m_alloc();

int  m_find();

int m_print();

int s_print();

void rp_format();

struct stack *s_alloc();

int s_clear();

int s_clearx();

int s_pop();

int s_push();

int s_swapxy();



/*? Clear memory*/

/*| cmem|err=m_clear(m1);|*/

int m_clear(me)

struct memory *me;

{  me->len=0; return 0;

}



struct memory *m_alloc(size, _er)

int size;

char *_er;

{  struct memory *foo;



   foo=malloc(sizeof(int)*2+sizeof(struct mem_element)*size);

   if(foo==NULL) strcpy(_er, "malloc");

   else

   { foo->len=0; foo->max=size;

   }

   return foo;

}



int m_find(m, var)

struct memory *m;

char *var;

{  int x,f;



   for(x=0, f=-1; x<m->len; x++)

      if(strcmp(var+2, m->e[x].name)==0) f=x;

   if(f==-1 && *var=='S')

   {  if(m->len >= m->max) return -1;

      else

      {  strcpy(m->e[m->len].name, var+2); f=m->len; m->len++; }

   } 

   return f;

}



/*? Show memory*/

/*| mem|err=m_print(m1,mode,fix,rd,erm);|*/

int m_print(m, form, fix, rd, er)

struct memory *m;

int form,fix;

double rd;

char *er;

{  int x;

   char buf[100];



   if(m->len==0) { strcpy(er,"memory empty"); return 1; }

   printf("\n");

   for(x=0; x<m->len; x++)

   {  printf("[%10s]  ", m->e[x].name);

      rp_format(m->e[x].d, buf, form, fix, rd);      

      printf("%s", buf);

   }

   return 0;

}



/*? Print stack*/

/*| print|printf("\n\n"); err=s_print(s1,mode,fix,rd,erm);|*/

int s_print(p, form, fix, rd, er)

struct stack *p;

int form,fix;

double rd;

char *er;

{  int x;

   char buf[100];



   if(p->len==0) { strcpy(er, "stack empty"); return 1; }

   for(x=p->len-1; x>=0; x--)

   {  rp_format(p->d[x], buf, form, fix, rd);

      printf("%s", buf);

   }

   return 0;

}



void rp_format(d, buf, form, fix, rd)

struct complex d;

char *buf;

int form, fix; 

double rd;

{  int y;

   double r,t;

   char p, n, bu[50], buu[100];



   ab2rt(d,&r,&t);

   if (d.b<0.0) {p='-'; d.b=-d.b;} else p='+';



   if(abs(form)==10)

      strcpy(bu, "% .0*f %cj %.0*f\t% .0*f < %.0*f\n");

   else

      strcpy(bu, "% .0*f %cj %.0*f\t%s\n");



   /* change %f format to %e if sci or r is too small to show with %f */

   if(form<0 || (r>0 && r<pow(10.0, -fix)))

      for(y=0; bu[y]!=0; y++) if(bu[y]=='f') bu[y]='e';



   /* change the precision digits in the format string */

   if(fix<10) n=48+fix;

   else /* replace 0 with 1 */

   {  n=48+fix-10;

      for(y=0; bu[y]!=0; y++) if(bu[y]=='0') bu[y]=49;

   }

   for(y=0; bu[y]!=0; y++) if(bu[y]=='*') bu[y]=n;



   if(abs(form)==10)

      sprintf(buf, bu, d.a, p, d.b, r, t*rd);

   else

   {  float2radix(buu, d.a, abs(form));

      sprintf(buf, bu, d.a, p, d.b, buu);

   }

}



struct stack *s_alloc(size, _er)

int size;

char *_er;

{  struct stack *foo;



   foo=malloc(sizeof(int)+sizeof(struct complex)*size);

   if(foo!=NULL) foo->len=size; else strcpy(_er, "malloc");

   return foo;

}



/*? Clear stack*/

/*| cs|err=s_clear(s1);|*/

int s_clear(st)

struct stack *st;

{  int x;



   for(x=0; x<st->len; x++) 

      st->d[x]=cx(0.0,0.0);

   return 0;

}



/*? Clear x*/

/*| cx|err=s_clearx(s1);|*/

int s_clearx(st)

struct stack *st;

{  st->d[0]=cx(0.0,0.0);

   return 0;

}



/*? Pop stack*/

/*| pop|err=s_pop(s1);|*/

int s_pop(st)

struct stack *st;

{  int x;



   for(x=0; x<st->len-1; x++)

   {  st->d[x].a=st->d[x+1].a; st->d[x].b=st->d[x+1].b;

   }

   st->d[st->len-1]=cx(0.0,0.0);

   return 0;

}



int s_push(st)

/*? Push stack*/

/*| push|err=s_push(s1);|*/

struct stack *st;

{  int x;



   for(x=st->len-1; x>0; x--)

   {  st->d[x].a=st->d[x-1].a; st->d[x].b=st->d[x-1].b;

   }

   return 0;

}



/*? Swap x and y on stack*/

/*| swap|err=s_swapxy(s1);|*/

int s_swapxy(st)

struct stack *st;

{  double a,b;



   a=st->d[0].a;          b=st->d[0].b;

   st->d[0].a=st->d[1].a; st->d[0].b=st->d[1].b;

   st->d[1].a=a;          st->d[1].b=b;

   return 0;

}

